# -*- mode: python; coding: utf-8 -*-
# Copyright 2017-2018 Peter Williams and collaborators.
# Licensed under the MIT License.

"""The main neural network code.

"""
from __future__ import absolute_import, division, print_function

__all__ = '''
NSModel
PhysicalApproximator
'''.split()

import os.path
from six.moves import range
import numpy as np
from pwkit import cgs
from pwkit.numutil import broadcastize
from keras import models, layers, optimizers

from . import DomainRange

hardcoded_nu_ref = 1.0
hardcoded_ne_ref = 1.0


class NSModel(models.Sequential):
    """Neuro-Synchro Model -- just keras.models.Sequential extended with some
    helpers specific to our data structures. If you run the `ns_setup` method
    you can train the neural net in our system.

    """
    # NOTE: __init__() must take no arguments in order for keras to be able to
    # deserialize NSModels from the HDF5 format.

    def ns_setup(self, result_index, data):
        self.result_index = int(result_index)
        self.domain_range = data.domain_range
        self.data = data
        assert self.result_index < self.domain_range.n_results
        return self # convenience


    def ns_fit(self, **kwargs):
        """Train this ANN model on the data in `self.data`. This function just
        takes care of extracting the right parameter and avoiding NaNs.

        """
        nres = self.data.norm_results[:,self.result_index]
        ok = np.isfinite(nres)
        nres = nres[ok].reshape((-1, 1))
        npar = self.data.norm_params[ok]
        return self.fit(npar, nres, **kwargs)


    def ns_validate(self, filter=True, to_phys=True):
        """Test this network by having it predict all of the values in our training
        sample. Returns `(params, actual, nn)`, where `params` is shape `(N,
        self.data.n_params)` and is the input parameters, `actual` is shape
        `(N,)` and is the actual values returned by the calculator, and `nn`
        is shape `(N,)` and is the values predicted by the neural net.

        If `filter` is true, the results will be filtered such that neither
        `actual` nor `nn` contain non-finite values.

        If `to_phys` is true, the values will be returned in the physical
        coordinate system. Otherwise they will be returned in the normalized
        coordinate system.

        """
        if to_phys:
            par = self.data.phys_params
            res = self.data.phys_results[:,self.result_index]
        else:
            par = self.data.norm_params
            res = self.data.norm_results[:,self.result_index]

        npred = self.predict(self.data.norm_params)[:,0]

        if filter:
            ok = np.isfinite(res) & np.isfinite(npred)
            par = par[ok]
            res = res[ok]
            npred = npred[ok]

        if to_phys:
            pred, _ = self.domain_range.rmaps[self.result_index].norm_to_phys(npred)
        else:
            pred = npred

        return par, res, pred


    def ns_sigma_clip(self, n_norm_sigma):
        """Assuming that self is already a decent approximation of the input data, try
        to improve things by NaN-ing out any measurements that are extremely
        discrepant with our approximation -- under the assumption that these
        are cases where the calculator went haywire.

        Note that this destructively modifies `self.data`.

        `n_norm_sigma` is the threshold above which discrepant values are
        flagged. It is evaluated using the differences between the neural net
        prediction and the training data in the *normalized* coordinate
        system.

        Returns the number of flagged points.

        """
        nres = self.data.norm_results[:,self.result_index]
        npred = self.predict(self.data.norm_params)[:,0]
        err = npred - nres
        m = np.nanmean(err)
        s = np.nanstd(err)
        bad = (np.abs((err - m) / s) > n_norm_sigma)
        self.data.phys[bad,self.domain_range.n_params+self.result_index] = np.nan
        self.data.norm[bad,self.domain_range.n_params+self.result_index] = np.nan
        return bad.sum()


    def ns_plot(self, param_index, plot_err=False, to_phys=False, thin=100):
        """Make a diagnostic plot comparing the approximation to the "actual" results
        from the calculator.

        """
        import omega as om

        par, act, nn = self.ns_validate(filter=True, to_phys=to_phys)

        if plot_err:
            err = nn - act
            p = om.quickXY(par[::thin,param_index], err[::thin], 'Error', lines=0)
        else:
            p = om.quickXY(par[::thin,param_index], act[::thin], 'Full calc', lines=0)
            p.addXY(par[::thin,param_index], nn[::thin], 'Neural', lines=0)

        return p


class PhysicalApproximator(object):
    """Approximate the eight nontrivial RT coefficients using a more
    physically-based parameterization.

    """
    results = 'j_I alpha_I rho_Q rho_V j_frac_pol alpha_frac_pol j_V_share alpha_V_share rho_Q_sign'.split()

    def __init__(self, nn_dir):
        self.domain_range = DomainRange.from_serialized(os.path.join(nn_dir, 'nn_config.toml'))

        for i, r in enumerate(self.results):
            assert self.domain_range.rmaps[i].name == r

            m = models.load_model(
                os.path.join(nn_dir, '%s.h5' % r),
                custom_objects = {'NSModel': NSModel}
            )
            m.result_index = i
            m.domain_range = self.domain_range
            setattr(self, r, m)


    @broadcastize(4, ret_spec=None)
    def compute_all_nontrivial(self, nu, B, n_e, theta, **kwargs):
        # Turn the standard parameters into the ones used in our computations

        no_B = ~(B > 0)
        nu_cyc = cgs.e * B / (2 * np.pi * cgs.me * cgs.c)
        nu_cyc[no_B] = 1e7 # fake to avoid div-by-0 for now
        kwargs['s'] = nu / nu_cyc

        # We are paranoid and assume that theta could take on any value ...
        # even though we do no bounds-checking for whether any of the *other*
        # inputs overlap the values that we used to train the neural nets.

        theta = theta % (2 * np.pi)
        w = (theta > np.pi)
        theta[w] = 2 * np.pi - theta[w]
        flip = (theta > 0.5 * np.pi)
        theta[flip] = np.pi - theta[flip]
        kwargs['theta'] = theta

        # Normalize inputs.

        oos_flags = 0

        norm = np.empty(nu.shape + (self.domain_range.n_params,))
        for i, mapping in enumerate(self.domain_range.pmaps):
            norm[...,i], flag = mapping.phys_to_norm(kwargs[mapping.name])
            if flag:
                oos_flags |= (1 << i)

        # Compute base outputs.

        j_I, flag = self.domain_range.rmaps[0].norm_to_phys(self.j_I.predict(norm)[...,0])
        if flag:
            oos_flags |= (1 << (self.domain_range.n_params + 0))

        alpha_I, flag = self.domain_range.rmaps[1].norm_to_phys(self.alpha_I.predict(norm)[...,0])
        if flag:
            oos_flags |= (1 << (self.domain_range.n_params + 1))

        rho_Q, flag = self.domain_range.rmaps[2].norm_to_phys(self.rho_Q.predict(norm)[...,0])
        if flag:
            oos_flags |= (1 << (self.domain_range.n_params + 2))

        rho_V, flag = self.domain_range.rmaps[3].norm_to_phys(self.rho_V.predict(norm)[...,0])
        if flag:
            oos_flags |= (1 << (self.domain_range.n_params + 3))

        j_frac_pol, flag = self.domain_range.rmaps[4].norm_to_phys(self.j_frac_pol.predict(norm)[...,0])
        if flag:
            oos_flags |= (1 << (self.domain_range.n_params + 4))

        alpha_frac_pol, flag = self.domain_range.rmaps[5].norm_to_phys(self.alpha_frac_pol.predict(norm)[...,0])
        if flag:
            oos_flags |= (1 << (self.domain_range.n_params + 5))

        j_V_share, flag = self.domain_range.rmaps[6].norm_to_phys(self.j_V_share.predict(norm)[...,0])
        if flag:
            oos_flags |= (1 << (self.domain_range.n_params + 6))

        alpha_V_share, flag = self.domain_range.rmaps[7].norm_to_phys(self.alpha_V_share.predict(norm)[...,0])
        if flag:
            oos_flags |= (1 << (self.domain_range.n_params + 7))

        rho_Q_sign, flag = self.domain_range.rmaps[8].norm_to_phys(self.rho_Q_sign.predict(norm)[...,0])
        if flag:
            oos_flags |= (1 << (self.domain_range.n_params + 8))

        # Patch up B = 0 in the obvious way. (Although if we ever have to deal
        # with nontrivial cold plasma densities, zones of zero B might affect
        # the RT if they cause refraction or what-have-you.)

        j_I[no_B] = 0.
        alpha_I[no_B] = 0.

        # Un-transform, baking in the invariant that our Q parameters are
        # always negative and the V parameters are always positive (given our
        # theta normalization).

        j_P = j_frac_pol * j_I
        j_V = j_V_share * j_P
        j_Q = -np.sqrt(1 - j_V_share**2) * j_P

        alpha_P = alpha_frac_pol * alpha_I
        alpha_V = alpha_V_share * alpha_P
        alpha_Q = -np.sqrt(1 - alpha_V_share**2) * alpha_P

        rho_Q = rho_Q * rho_Q_sign

        # Now apply the known scalings.

        n_e_scale = n_e / hardcoded_ne_ref
        j_I *= n_e_scale
        alpha_I *= n_e_scale
        j_Q *= n_e_scale
        alpha_Q *= n_e_scale
        j_V *= n_e_scale
        alpha_V *= n_e_scale
        rho_Q *= n_e_scale
        rho_V *= n_e_scale

        freq_scale = nu / hardcoded_nu_ref
        j_I *= freq_scale
        alpha_I /= freq_scale
        j_Q *= freq_scale
        alpha_Q /= freq_scale
        j_V *= freq_scale
        alpha_V /= freq_scale
        rho_Q /= freq_scale
        rho_V /= freq_scale

        theta_sign_term = np.ones(n_e.shape, dtype=np.int)
        theta_sign_term[flip] = -1
        j_V *= theta_sign_term
        alpha_V *= theta_sign_term
        rho_V *= theta_sign_term

        # Pack it up and we're done.

        result = np.empty(n_e.shape + (8,))
        result[:,0] = j_I
        result[:,1] = alpha_I
        result[:,2] = j_Q
        result[:,3] = alpha_Q
        result[:,4] = j_V
        result[:,5] = alpha_V
        result[:,6] = rho_Q
        result[:,7] = rho_V
        return result, oos_flags



# -*- mode: python; coding: utf-8 -*-
# Copyright 2017-2018 Peter Williams and collaborators.
# Licensed under the MIT License.

"""Train and use artificial neural network approximations ("regressions") of
synchrotron radiative transfer coefficients as a function of various physical
input parameters.

The meat of the neural network code is in the ``impl`` module to avoid
importing Keras unless needed.

"""
from __future__ import absolute_import, division, print_function

__all__ = '''
AbsLogMapping
DirectMapping
DomainRange
LogMapping
LogitMapping
Mapping
NegLogMapping
NinthRootMapping
SampleData
SignMapping
basic_load
mapping_from_dict
mapping_from_samples
'''.split()

from collections import OrderedDict
from six.moves import range
import numpy as np
import pandas as pd
from pwkit.io import Path


class Mapping(object):
    desc = None # set by subclasses
    trainer = None
    phys_bounds_mode = 'empirical'
    normalization_mode = 'gaussian'
    out_of_sample = 'ignore'

    def __init__(self, name):
        self.name = name


    @classmethod
    def from_info_and_samples(cls, info, phys_samples):
        inst = cls.from_dict(info, load_bounds=False)

        valid = np.isfinite(phys_samples) & inst._is_valid(phys_samples)
        n_rej = phys_samples.size - valid.sum()
        print('%s: rejecting %d samples out of %d' % (inst.name, n_rej, phys_samples.size))
        phys_samples = phys_samples[valid]
        if phys_samples.size < 3:
            raise Exception('not enough valid samples for %s' % inst.name)

        if inst.phys_bounds_mode == 'empirical':
            inst.p_min = phys_samples.min()
            inst.p_max = phys_samples.max()
        elif inst.phys_bounds_mode == 'theta':
            inst.p_min = 0.
            inst.p_max = 0.5 * np.pi
        else:
            raise ValueError('unrecognized phys_bounds_mode value %r for %s' %
                             (inst.phys_bounds_mode, inst.name))

        # Pluggable "transform"
        transformed = inst._to_xform(phys_samples)

        if inst.normalization_mode == 'gaussian':
            inst.x_mean = transformed.mean()
            inst.x_std = transformed.std()
        elif inst.normalization_mode == 'unit_interval':
            # Maps the physical values to the unit interval [0, 1].
            inst.x_mean = transformed.min()
            inst.x_std = transformed.max() - inst.x_mean
        else:
            raise ValueError('unrecognized normalization_mode value %r for %s' %
                             (inst.normalization_mode, inst.name))

        # Normalize
        normed = (transformed - inst.x_mean) / inst.x_std
        inst.n_min = normed.min()
        inst.n_max = normed.max()

        return inst


    def __repr__(self):
        return '<Mapping %s %s mean=%r sd=%r>' % (self.name, self.desc, self.x_mean, self.x_std)


    def phys_to_norm(self, phys):
        # note: using prefix ~ instead of np.logical_not fails for scalars
        oos = np.logical_not((phys >= self.p_min) & (phys <= self.p_max)) # catches NaNs
        any_oos = np.any(oos)

        if any_oos:
            if self.out_of_sample == 'ignore':
                pass
            elif self.out_of_sample == 'clip':
                phys = np.clip(phys, self.p_min, self.p_max)
            elif self.out_of_sample == 'nan':
                phys = phys.copy()
                phys[oos] = np.nan
            else:
                raise Exception('unrecognized out-of-sample behavior %r' % self.out_of_sample)

        return (self._to_xform(phys) - self.x_mean) / self.x_std, any_oos


    def norm_to_phys(self, norm):
        oos = np.logical_not((norm >= self.n_min) & (norm <= self.n_max)) # catches NaNs
        any_oos = np.any(oos)

        if any_oos:
            if self.out_of_sample == 'ignore':
                pass
            elif self.out_of_sample == 'clip':
                norm = np.clip(norm, self.n_min, self.n_max)
            elif self.out_of_sample == 'nan':
                norm = norm.copy()
                norm[oos] = np.nan
            else:
                raise Exception('unrecognized out-of-sample behavior %r' % self.out_of_sample)

        return self._from_xform(norm * self.x_std + self.x_mean), any_oos


    def to_dict(self):
        d = OrderedDict()
        d['name'] = self.name
        d['maptype'] = self.desc

        if self.phys_bounds_mode is not None:
            d['phys_bounds_mode'] = self.phys_bounds_mode
        if self.normalization_mode is not None:
            d['normalization_mode'] = self.normalization_mode
        if self.trainer is not None:
            d['trainer'] = self.trainer
        if self.out_of_sample is not None:
            d['out_of_sample'] = self.out_of_sample

        d['x_mean'] = self.x_mean
        d['x_std'] = self.x_std
        d['phys_min'] = self.p_min
        d['phys_max'] = self.p_max
        d['norm_min'] = self.n_min
        d['norm_max'] = self.n_max
        return d


    @classmethod
    def from_dict(cls, info, load_bounds=True):
        if str(info['maptype']) != cls.desc:
            raise ValueError('info is for maptype %s but this class is %s' % (info['maptype'], cls.desc))

        inst = cls(str(info['name']))
        if 'phys_bounds_mode' in info:
            inst.phys_bounds_mode = info['phys_bounds_mode']
        if 'normalization_mode' in info:
            inst.normalization_mode = info['normalization_mode']
        if 'trainer' in info:
            inst.trainer = info['trainer']
        if 'out_of_sample' in info:
            inst.out_of_sample = info['out_of_sample']

        if load_bounds:
            inst.x_mean = float(info['x_mean'])
            inst.x_std = float(info['x_std'])
            inst.p_min = float(info['phys_min'])
            inst.p_max = float(info['phys_max'])
            inst.n_min = float(info['norm_min'])
            inst.n_max = float(info['norm_max'])

        return inst


class AbsLogMapping(Mapping):
    desc = 'abs_log'

    def _to_xform(self, p):
        return np.log10(np.abs(p))

    def _from_xform(self, x):
        return 10**x # XXX not invertible!

    def _is_valid(self, p):
        return p != 0


class DirectMapping(Mapping):
    desc = 'direct'

    def _to_xform(self, p):
        return p

    def _from_xform(self, x):
        return x

    def _is_valid(self, p):
        return np.ones(p.shape, dtype=np.bool)


class LogMapping(Mapping):
    desc = 'log'

    def _to_xform(self, p):
        return np.log10(p)

    def _from_xform(self, x):
        return 10**x

    def _is_valid(self, p):
        return (p > 0)


class LogitMapping(Mapping):
    desc = 'logit'

    def _to_xform(self, p):
        return np.log(p / (1. - p))

    def _from_xform(self, x):
        return np.exp(x) / (np.exp(x) + 1)

    def _is_valid(self, p):
        # Infinities are hard to deal with so we don't allow p = 0 or p = 1.
        return (p > 0) & (p < 1)


class NegLogMapping(Mapping):
    desc = 'neg_log'

    def _to_xform(self, p):
        return np.log10(-p)

    def _from_xform(self, x):
        return -(10**x)

    def _is_valid(self, p):
        return (p < 0)


class NinthRootMapping(Mapping):
    desc = 'ninth_root'

    def _to_xform(self, p):
        return np.cbrt(np.cbrt(p))

    def _from_xform(self, x):
        return x**9

    def _is_valid(self, p):
        return np.ones(p.shape, dtype=np.bool)


class SignMapping(Mapping):
    desc = 'sign'

    def _to_xform(self, p):
        return np.sign(p)

    def _from_xform(self, x):
        return np.sign(x) # XXX not reversible!

    def _is_valid(self, p):
        return (p != 0.)


_mappings = {
    'abs_log': AbsLogMapping,
    'direct': DirectMapping,
    'log': LogMapping,
    'logit': LogitMapping,
    'neg_log': NegLogMapping,
    'ninth_root': NinthRootMapping,
    'sign': SignMapping,
}


class Passthrough(object):
    def __init__(self, d):
        self.d = d

    def to_dict(self):
        return self.d


def mapping_from_info_and_samples(info, phys_samples):
    cls = _mappings[info['maptype']]
    return cls.from_info_and_samples(info, phys_samples)

def mapping_from_dict(info):
    maptype = str(info['maptype'])
    cls = _mappings[maptype]
    return cls.from_dict(info)


def basic_load(datadir, drop_metadata=True):
    datadir = Path(datadir)
    chunks = []

    for item in datadir.glob('*.txt'):
        chunks.append(pd.read_table(str(item)))

    data = pd.concat(chunks, ignore_index=True)

    if drop_metadata:
        # Drop `foo(mtea)` columns
        for col in data.columns:
            if col.endswith('(meta)'):
                del data[col]

    return data


class DomainRange(object):
    n_params = None
    n_results = None
    pmaps = None
    rmaps = None


    @classmethod
    def from_info_and_samples(cls, info, df):
        inst = cls()
        inst.n_params = len(info['params'])
        inst.n_results = len(info['results'])
        inst.pmaps = []
        inst.rmaps = []

        for i, pinfo in enumerate(info['params']):
            if not df.columns[i].startswith(pinfo['name'] + '('):
                raise Exception('alignment error: expect data column %d to be %r; got %r' % (i, pinfo['name'], df.columns[i]))
            inst.pmaps.append(mapping_from_info_and_samples(pinfo, df[df.columns[i]].values))

        n = inst.n_params

        for i, rinfo in enumerate(info['results']):
            if not df.columns[n+i].startswith(rinfo['name'] + '(res)'):
                raise Exception('alignment error: expect data column %d to be %r; got %r' % (n + i, rinfo['name'], df.columns[n + i]))
            inst.rmaps.append(mapping_from_info_and_samples(rinfo, df[df.columns[n+i]].values))

        return inst


    @classmethod
    def from_serialized(cls, config_path, result_to_extract=None):
        """`result_to_extract` is a total lazy hack for the training tool."""
        import pytoml

        with Path(config_path).open('rt') as f:
            info = pytoml.load(f)

        inst = cls()
        inst.pmaps = []
        inst.rmaps = []
        extracted_info = None

        for subinfo in info['params']:
            inst.pmaps.append(mapping_from_dict(subinfo))

        for i, subinfo in enumerate(info['results']):
            if result_to_extract is not None and subinfo['name'] == result_to_extract:
                extracted_info = subinfo
                extracted_info['_index'] = i
            inst.rmaps.append(mapping_from_dict(subinfo))

        inst.n_params = len(inst.pmaps)
        inst.n_results = len(inst.rmaps)

        if result_to_extract is not None:
            return inst, extracted_info
        return inst


    def __repr__(self):
        return '\n'.join(
            ['<%s n_p=%d n_r=%d' % (self.__class__.__name__, self.n_params, self.n_results)] +
            ['  P%d=%r,' % (i, m) for i, m in enumerate(self.pmaps)] +
            ['  R%d=%r,' % (i, m) for i, m in enumerate(self.rmaps)] +
            ['>'])


    def into_info(self, info):
        info['params'] = [m.to_dict() for m in self.pmaps]
        info['results'] = [m.to_dict() for m in self.rmaps]


    def load_and_normalize(self, datadir):
        df = basic_load(datadir)

        for i, pmap in enumerate(self.pmaps):
            if not df.columns[i].startswith(pmap.name + '('):
                raise Exception('alignment error: expect data column %d to be %s(...); got %s' % (i, pmap.name, df.columns[i]))

        n = self.n_params

        for i, rmap in enumerate(self.rmaps):
            if df.columns[n+i] != rmap.name + '(res)':
                raise Exception('alignment error: expect data column %d to be %s(res); got %s' % (n + i, rmap.name, df.columns[n + i]))

        return SampleData(self, df)


class SampleData(object):
    df = None
    domain_range = None
    phys = None
    norm = None
    oos_flags = 0

    def __init__(self, domain_range, df):
        self.df = df
        self.domain_range = domain_range
        self.phys = df.values
        self.norm = np.empty_like(self.phys)

        for i in range(self.domain_range.n_params):
            self.norm[:,i], flag = self.domain_range.pmaps[i].phys_to_norm(self.phys[:,i])
            if flag:
                self.oos_flags |= (1 << i)

        for i in range(self.domain_range.n_results):
            j = i + self.domain_range.n_params
            self.norm[:,j], flag = self.domain_range.rmaps[i].phys_to_norm(self.phys[:,j])
            if flag:
                self.oos_flags |= (1 << j)

    @property
    def phys_params(self):
        return self.phys[:,:self.domain_range.n_params]

    @property
    def phys_results(self):
        return self.phys[:,self.domain_range.n_params:]

    @property
    def norm_params(self):
        return self.norm[:,:self.domain_range.n_params]

    @property
    def norm_results(self):
        return self.norm[:,self.domain_range.n_params:]



#! /usr/bin/env python
# -*- mode: python; coding: utf-8 -*-
# Copyright 2017-2018 Peter Williams and collaborators.
# Licensed under the MIT License.

"""Command-line access to neurosynchro functionality.

"""
from __future__ import absolute_import, division, print_function

import argparse, sys
import numpy as np
from pwkit.cli import die
from pwkit.io import Path
import pytoml

from . import basic_load


# The "lock-domain-range" subcommand

def _hack_pytoml():
    """pytoml will stringify floats using repr, which is ugly and fails outright with
    very small values (i.e. 1e-30 becomes "0.000...."). Here we hack it to use
    exponential notation if needed.

    """
    from pytoml import writer
    orig_format_value = writer._format_value

    if not getattr(orig_format_value, '_neurosynchro_hack_applied', False):
        def better_format_value(v):
            if isinstance(v, float):
                if not np.isfinite(v):
                    raise ValueError("{0} is not a valid TOML value".format(v))
                return '%.16g' % v
            return orig_format_value(v)

        better_format_value._neurosynchro_hack_applied = True
        writer._format_value = better_format_value

_hack_pytoml()


def make_ldr_parser():
    ap = argparse.ArgumentParser()
    ap.add_argument('datadir', type=str, metavar='DATADIR',
                    help='The path to the input sample data directory.')
    ap.add_argument('nndir', type=str, metavar='NNDIR',
                    help='The path to the output neural-net directory.')
    return ap


def lock_domain_range_cli(args):
    from . import DomainRange

    settings = make_ldr_parser().parse_args(args=args)

    # Load samples
    df = basic_load(settings.datadir)

    # Load skeleton config
    cfg_path = Path(settings.nndir) / 'nn_config.toml'
    with cfg_path.open('rt') as f:
        info = pytoml.load(f)

    # Turn into processed DomainRange object
    dr = DomainRange.from_info_and_samples(info, df)

    # Update config and rewrite
    dr.into_info(info)

    with cfg_path.open('wt') as f:
        pytoml.dump(f, info)


# The "summarize" subcommand

def summarize(datadir):
    df = basic_load(datadir)

    # Report stuff.

    print('Columns:', ' '.join(df.columns))
    print('Number of rows:', df.shape[0])
    print('Total number of NaNs:', np.isnan(df.values).sum())
    print('Number of rows with NaNs:', (np.isnan(df.values).sum(axis=1) > 0).sum())

    for c in df.columns:
        r = df[c]
        print()
        print('Column %s:' % c)
        print('  Number of NaNs:', np.isnan(r).sum())
        print('  Non-NaN max:', np.nanmax(r))
        print('  Non-NaN min:', np.nanmin(r))
        print('  Nonnegative:', (r >= 0).sum())
        print('  Nonpositive:', (r <= 0).sum())


def make_summarize_parser():
    ap = argparse.ArgumentParser()
    ap.add_argument('datadir', type=str, metavar='DATADIR',
                    help='The path to the sample data directory.')
    return ap


def summarize_cli(args):
    settings = make_summarize_parser().parse_args(args=args)
    summarize(settings.datadir)


# The "transform" subcommand

def transform(datadir):
    """This task takes the raw synchrotron coefficients output by rimphony and
    transforms them into a format that better respects the physical
    constraints of the problem.

    """
    import pandas as pd

    df = basic_load(datadir)
    n = df.shape[0]

    df = df.dropna()
    print('Dropping due to NaNs:', n - df.shape[0], file=sys.stderr)

    bad = (df['j_I(res)'] <= 0)
    mask = bad
    print('Rows with bad J_I:', bad.sum(), file=sys.stderr)

    bad = (df['alpha_I(res)'] <= 0)
    mask |= bad
    print('Rows with bad a_I:', bad.sum(), file=sys.stderr)

    bad = (df['j_Q(res)'] >= 0)
    mask |= bad
    print('Rows with bad J_Q:', bad.sum(), file=sys.stderr)

    bad = (df['alpha_Q(res)'] >= 0)
    mask |= bad
    print('Rows with bad a_Q:', bad.sum(), file=sys.stderr)

    bad = (df['j_V(res)'] <= 0)
    mask |= bad
    print('Rows with bad J_V:', bad.sum(), file=sys.stderr)

    bad = (df['alpha_V(res)'] <= 0)
    mask |= bad
    print('Rows with bad a_V:', bad.sum(), file=sys.stderr)

    # This cut isn't physically motivated, but under the current rimphony
    # model, f_V is always positive.
    bad = (df['rho_V(res)'] <= 0)
    mask |= bad
    print('Rows with bad f_V:', bad.sum(), file=sys.stderr)

    n = df.shape[0]
    df = df[~mask]
    print('Dropped due to first-pass filters:', n - df.shape[0], file=sys.stderr)

    j_pol = np.sqrt(df['j_Q(res)']**2 + df['j_V(res)']**2)
    a_pol = np.sqrt(df['alpha_Q(res)']**2 + df['alpha_V(res)']**2)

    df['j_frac_pol(res)'] = j_pol / df['j_I(res)']
    bad = (df['j_frac_pol(res)'] < 0) | (df['j_frac_pol(res)'] > 1)
    mask = bad
    print('Rows with bad j_frac_pol:', bad.sum(), file=sys.stderr)

    df['alpha_frac_pol(res)'] = a_pol / df['alpha_I(res)']
    bad = (df['alpha_frac_pol(res)'] < 0) | (df['alpha_frac_pol(res)'] > 1)
    mask |= bad
    print('Rows with bad alpha_frac_pol:', bad.sum(), file=sys.stderr)

    n = df.shape[0]
    df = df[~mask]
    print('Dropped due to second-pass filters:', n - df.shape[0], file=sys.stderr)

    df['j_V_share(res)'] = df['j_V(res)'] / j_pol
    df['alpha_V_share(res)'] = df['alpha_V(res)'] / a_pol

    # I used to scale rho_{Q,V} by alpha_I, but these values are often
    # strongly different. (And, judging by the commentary in Heyvaerts, I
    # think this is probably OK and not a sign of a numerics problem.) So we
    # just pass those columns on through like {j,alpha}_I. However, in a bit
    # of a hack, we add a column giving the sign of the rho_Q column, since in
    # the "pitchy kappa" distribution we have a non-negligible number of
    # negative rho_Q values *plus* a large dynamic range on both sides of
    # zero. Adding this column lets us break the neural networking into two
    # pieces in a way that doesn't involve a bunch of complicated
    # rearchitecting of my parameter code.
    df['rho_Q_sign(res)'] = np.sign(df['rho_Q(res)'])

    print('Final row count:', df.shape[0], file=sys.stderr)

    for c in 'j_Q alpha_Q j_V alpha_V'.split():
        del df[c + '(res)']

    df.to_csv(
        sys.stdout,
        sep = '\t',
        index = False
    )


def make_transform_parser():
    ap = argparse.ArgumentParser()
    ap.add_argument('datadir', type=str, metavar='DATADIR',
                    help='The path to the sample data directory.')
    return ap


def transform_cli(args):
    settings = make_transform_parser().parse_args(args=args)
    transform(settings.datadir)


# The entrypoint

def entrypoint(argv):
    if len(argv) == 1:
        die('must supply a subcommand: "lock-domain-range", "summarize", "train", "transform"')

    if argv[1] == 'lock-domain-range':
        lock_domain_range_cli(argv[2:])
    elif argv[1] == 'summarize':
        summarize_cli(argv[2:])
    elif argv[1] == 'train':
        from .train import train_cli
        train_cli(argv[2:])
    elif argv[1] == 'transform':
        transform_cli(argv[2:])
    else:
        die('unrecognized subcommand %r', argv[1])


def main():
    import sys
    from pwkit import cli

    cli.unicode_stdio()
    cli.propagate_sigint()
    cli.backtrace_on_usr1()
    entrypoint(sys.argv)
